@@ -18,11 +18,19 @@ from abc import abstractmethod
 import copy
 import io
 import json
+from logging import getLogger
+from logging import Logger
 import os
 import re
 from typing import Dict
+try:  # typing.final is not available in python < 3.8 so create a dummy decorator in those cases
+    from typing import final
+except ImportError:
+    def final(meth):
+        return meth
 from typing import List
 from typing import Optional
+from typing import Set
 
 from entrypoints import get_group_all
 from jsonschema import draft7_format_checker
@@ -42,10 +50,22 @@ class SchemaManager(SingletonConfigurable):
     """Singleton used to store all schemas for all metadata types.
        Note: we currently don't refresh these entries.
     """
+    # Maps SchemaspaceID AND SchemaspaceName to Schemaspace instance (two entries from Schemaspace instance)
+    schemaspaces: Dict[str, 'Schemaspace']
+
+    # Maps SchemaspaceID to ShemaspaceName
+    schemaspace_id_to_name: Dict[str, str]
+
+    # Maps SchemaspaceID to mapping of schema name to SchemasProvider
+    schemaspace_schemasproviders: Dict[str, Dict[str, 'SchemasProvider']]
 
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
 
+        self.schemaspaces = {}
+        self.schemaspace_id_to_name = {}
+        self.schemaspace_schemasproviders = {}
+
         # The following exposes the metadata-test schemaspace if true or 1.
         # Metadata testing will enable this env.  Note: this cannot be globally
         # defined, else the file could be loaded before the tests have enable the env.
@@ -57,9 +77,18 @@ class SchemaManager(SingletonConfigurable):
             self._meta_schema = json.load(f)
         self._load_schemaspace_schemas()
 
-    def get_schemaspace_names(self) -> list:
+    def get_schemaspace_names(self, include_deprecated: bool = False) -> List[str]:
         """Returns list of registered schemaspace names."""
-        return list(self.schemaspace_id_to_name.values())
+
+        schemaspace_names: List[str] = []
+        # Filter out deprecated schemaspaces
+        for id, name in self.schemaspace_id_to_name.items():
+            if not include_deprecated and self.schemaspaces.get(id).is_deprecated:
+                self.log.debug(f"get_schemaspace_names: skipping deprecated schemaspace '{name}'...")
+                continue
+            schemaspace_names.append(name)
+
+        return schemaspace_names
 
     def get_schemaspace_name(self, schemaspace_name_or_id: str) -> str:
         """Returns the human-readable name of the given schemaspace name or id.
@@ -73,8 +102,11 @@ class SchemaManager(SingletonConfigurable):
     def get_schemaspace_schemas(self, schemaspace_name_or_id: str) -> dict:
         self._validate_schemaspace(schemaspace_name_or_id)
         schemaspace = self.schemaspaces.get(schemaspace_name_or_id.lower())
-        schemas = schemaspace.schemas
-        return copy.deepcopy(schemas)
+        schemas = copy.deepcopy(schemaspace.schemas)
+        # Filter out deprecated schemas
+        for schema_name in schemaspace.deprecated_schema_names:
+            del schemas[schema_name]
+        return schemas
 
     def get_schema(self, schemaspace_name_or_id: str, schema_name: str) -> dict:
         """Returns the specified schema for the specified schemaspace."""
@@ -90,7 +122,12 @@ class SchemaManager(SingletonConfigurable):
     def get_schemaspace(self, schemaspace_name_or_id: str) -> 'Schemaspace':
         """Returns the Schemaspace instance associated with the given name or id."""
         self._validate_schemaspace(schemaspace_name_or_id)
-        return copy.deepcopy(self.schemaspaces.get(schemaspace_name_or_id.lower()))
+        schemaspace = self.schemaspaces.get(schemaspace_name_or_id.lower())
+        return copy.deepcopy(schemaspace)
+
+    def get_schemasproviders(self, schemaspace_id: str) -> Dict[str, 'SchemasProvider']:
+        """Returns a dictionary of schema name to SchemasProvider instance within a given schemaspace. """
+        return self.schemaspace_schemasproviders.get(schemaspace_id, {})
 
     def clear_all(self) -> None:
         """Primarily used for testing, this method reloads schemas from initial values. """
@@ -105,8 +142,7 @@ class SchemaManager(SingletonConfigurable):
 
     def _load_schemaspace_schemas(self):
         """Gets Schemaspaces and SchemasProviders via entrypoints and validates/loads their schemas."""
-        self.schemaspaces: Dict[str, Schemaspace] = {}
-        self.schemaspace_id_to_name: Dict[str, str] = {}
+
         self._load_schemaspaces()
         self._load_schemas_providers()
         # Issue a warning for any "empty" schemaspaces...
@@ -159,28 +195,33 @@ class SchemaManager(SingletonConfigurable):
                 schemas = schemas_provider.get_schemas()
                 for schema in schemas:
                     try:
-                        schemaspace_id = schema.get("schemaspace_id")
+                        schemaspace_id = schema.get("schemaspace_id").lower()
                         schemaspace_name = schema.get("schemaspace")
                         schema_name = schema.get("name")
                         # Ensure that both schemaspace id and name are registered and both point to same instance
-                        if schemaspace_id.lower() not in self.schemaspaces:
+                        if schemaspace_id not in self.schemaspaces:
                             raise ValueError(f"Schema '{schema_name}' references a schemaspace "
                                              f"'{schemaspace_id}' that is not loaded!")
-                        if schemaspace_name.lower() not in self.schemaspaces:
+                        if schemaspace_name not in self.schemaspaces:
                             raise ValueError(f"Schema '{schema_name}' references a schemaspace "
                                              f"'{schemaspace_name}' that is not loaded!")
-                        if self.schemaspaces[schemaspace_id.lower()] != self.schemaspaces[schemaspace_name.lower()]:
+                        if self.schemaspaces[schemaspace_id] != self.schemaspaces[schemaspace_name.lower()]:
                             raise ValueError(f"Schema '{schema_name}' references a schemaspace name "
                                              f"'{schemaspace_name}' and a schemaspace id '{schemaspace_id}' "
                                              f"that are associated with different Schemaspace instances!")
 
+                        schema = self.schemaspaces[schemaspace_id].filter_schema(schema)
                         self._validate_schema(schemaspace_name, schema_name, schema)
                         # Only add the schema once since schemaspace_name is pointing to the same Schemaspace instance.
-                        self.schemaspaces[schemaspace_id.lower()].add_schema(schema)
+                        self.schemaspaces[schemaspace_id].add_schema(schema)
+                        providers = self.schemaspace_schemasproviders.get(schemaspace_id, {})
+                        providers[schema_name] = schemas_provider  # Capture the schemasprovider for this schema
+                        self.schemaspace_schemasproviders[schemaspace_id] = providers
 
                     except Exception as schema_err:
                         self.log.error(f"Error loading schema '{schema.get('name', '??')}' for SchemasProvider "
                                        f"'{schemas_provider_ep.name}' - {schema_err}")
+
             except Exception as provider_err:
                 # log and ignore initialization errors
                 self.log.error(f"Error loading schemas for SchemasProvider "
@@ -216,6 +257,8 @@ class Schemaspace(LoggingConfigurable):
     _display_name: str
     _description: str
     _schemas: Dict[str, Dict]  # use a dict to prevent duplicate entries
+    _deprecated: bool
+    _deprecated_schema_names: Set[str]
 
     def __init__(self,
                  schemaspace_id: str,
@@ -226,6 +269,8 @@ class Schemaspace(LoggingConfigurable):
         super().__init__(**kwargs)
 
         self._schemas = {}
+        self._deprecated = False
+        self._deprecated_schema_names = set()
 
         # Validate properties
         if not schemaspace_id:
@@ -270,10 +315,45 @@ class Schemaspace(LoggingConfigurable):
         """Returns the schemas currently associated with this schemaspace"""
         return self._schemas
 
+    @property
+    def is_deprecated(self) -> bool:
+        """Indicates if this schemaspace is deprecaated"""
+        return self._deprecated
+
+    @property
+    def deprecated_schema_names(self) -> List[str]:
+        """Returns a list of deprecated schema names associated with this schemaspace"""
+        return list(self._deprecated_schema_names)
+
+    def filter_schema(self, schema: Dict) -> Dict:
+        """Allows Schemaspace to apply changes to a given schema prior to its validation (and add)."""
+        return schema
+
     def add_schema(self, schema: Dict) -> None:
         """Associates the given schema to this schemaspace"""
         assert isinstance(schema, dict), "Parameter 'schema' is not a dictionary!"
         self._schemas[schema.get('name')] = schema
+        if schema.get('deprecated', False):
+            self._deprecated_schema_names.add(schema.get('name'))
+
+    @final
+    def migrate(self, *args, **kwargs) -> List[str]:
+        """Migrate schemaspace instances.  This method is `final` and should not be overridden.
+
+        Its purpose is to drive migration across the Schemaspace's SchemasProviders and gather
+        results.
+
+        Returns the list of migrated instance names.
+        """
+        migrated_instances: List[str] = []
+        # For each schema in this schemaspace, invoke its corresponding
+        # SchemasProvider's migrate method and collect the results.
+        schema_to_provider = SchemaManager.instance().get_schemasproviders(self.id)
+        for schema_name, provider in schema_to_provider.items():
+            instances = provider.migrate(schemaspace_name=self.name, schema_name=schema_name)
+            migrated_instances.extend(instances)
+
+        return migrated_instances
 
     @staticmethod
     def _validate_id(id) -> bool:
@@ -299,7 +379,25 @@ class Schemaspace(LoggingConfigurable):
 class SchemasProvider(ABC):
     """Abstract base class used to obtain schema definitions from registered schema providers."""
 
+    log: Logger
+
+    def __init__(self, *args, **kwargs):
+        self.log = getLogger('ElyraApp')
+
     @abstractmethod
     def get_schemas(self) -> List[Dict]:
         """Returns a list of schemas"""
         pass
+
+    def migrate(self, *args, **kwargs) -> List[str]:
+        """Migrate instances of schemas provided by this SchemasProvider.
+
+        kwargs:
+        schema_name: str  The name of the schema
+        schemaspace_name: str  The name of the schemaspace in which this schema resides
+
+        Called by Schemaspace.migrate(), this method will be called with a `schema_name`
+        keyword argument indicating a name of schema provided by this SchemasProvider.
+        The method will return a list of migrated instances or an empty array.
+        """
+        return list()
